@@ -13,8 +13,8 @@ from hashlib import sha256
 from threading import Thread
 
 from PyQt5.Qt import (
-    QApplication, QDockWidget, QEvent, QMimeData, QModelIndex, QPixmap, QScrollBar,
-    Qt, QToolBar, QUrl, QVBoxLayout, QWidget, pyqtSignal
+    QApplication, QCursor, QDockWidget, QEvent, QMenu, QMimeData, QModelIndex,
+    QPixmap, Qt, QTimer, QToolBar, QUrl, QVBoxLayout, QWidget, pyqtSignal
 )
 
 from calibre import prints
@@ -25,12 +25,14 @@ from calibre.gui2.dialogs.drm_error import DRMErrorMessage
 from calibre.gui2.image_popup import ImagePopup
 from calibre.gui2.main_window import MainWindow
 from calibre.gui2.viewer.annotations import (
-    merge_annotations, parse_annotations, save_annots_to_epub, serialize_annotations
+    merge_annotations, parse_annotations, save_annots_to_epub, serialize_annotation,
+    serialize_annotations
 )
 from calibre.gui2.viewer.bookmarks import BookmarkManager
 from calibre.gui2.viewer.convert_book import (
     clean_running_workers, prepare_book, update_book
 )
+from calibre.gui2.viewer.highlights import HighlightsPanel
 from calibre.gui2.viewer.lookup import Lookup
 from calibre.gui2.viewer.overlay import LoadingOverlay
 from calibre.gui2.viewer.search import SearchPanel
@@ -43,6 +45,7 @@ from calibre.gui2.viewer.web_view import (
 from calibre.utils.date import utcnow
 from calibre.utils.img import image_from_path
 from calibre.utils.ipc.simple_worker import WorkerError
+from calibre.utils.iso8601 import parse_iso8601
 from calibre.utils.monotonic import monotonic
 from calibre.utils.serialize import json_loads
 from polyglot.builtins import as_bytes, iteritems, itervalues
@@ -71,6 +74,7 @@ def dock_defs():
     d(_('Bookmarks'), 'bookmarks', Qt.RightDockWidgetArea)
     d(_('Search'), 'search', Qt.LeftDockWidgetArea)
     d(_('Inspector'), 'inspector', Qt.RightDockWidgetArea, Qt.AllDockWidgetAreas)
+    d(_('Highlights'), 'highlights', Qt.RightDockWidgetArea)
     return ans
 
 
@@ -78,13 +82,6 @@ def path_key(path):
     return sha256(as_bytes(path)).hexdigest()
 
 
-class ScrollBar(QScrollBar):
-
-    def paintEvent(self, ev):
-        if self.isEnabled():
-            return QScrollBar.paintEvent(self, ev)
-
-
 class EbookViewer(MainWindow):
 
     msg_from_anotherinstance = pyqtSignal(object)
@@ -94,11 +91,14 @@ class EbookViewer(MainWindow):
 
     def __init__(self, open_at=None, continue_reading=None, force_reload=False):
         MainWindow.__init__(self, None)
-        self.shutting_down = False
+        self.shutting_down = self.close_forced = False
         self.force_reload = force_reload
         connect_lambda(self.book_preparation_started, self, lambda self: self.loading_overlay(_(
             'Preparing book for first read, please wait')), type=Qt.QueuedConnection)
         self.maximized_at_last_fullscreen = False
+        self.save_pos_timer = t = QTimer(self)
+        t.setSingleShot(True), t.setInterval(3000), t.setTimerType(Qt.VeryCoarseTimer)
+        connect_lambda(t.timeout, self, lambda self: self.save_annotations(in_book_file=False))
         self.pending_open_at = open_at
         self.base_window_title = _('E-book viewer')
         self.setWindowTitle(self.base_window_title)
@@ -154,6 +154,9 @@ class EbookViewer(MainWindow):
         w.toggle_requested.connect(self.toggle_bookmarks)
         self.bookmarks_dock.setWidget(w)
 
+        self.highlights_widget = w = HighlightsPanel(self)
+        self.highlights_dock.setWidget(w)
+
         self.web_view = WebView(self)
         self.web_view.cfi_changed.connect(self.cfi_changed)
         self.web_view.reload_book.connect(self.reload_book)
@@ -163,6 +166,8 @@ class EbookViewer(MainWindow):
         self.search_widget.show_search_result.connect(self.web_view.show_search_result)
         self.web_view.search_result_not_found.connect(self.search_widget.search_result_not_found)
         self.web_view.toggle_bookmarks.connect(self.toggle_bookmarks)
+        self.web_view.toggle_highlights.connect(self.toggle_highlights)
+        self.web_view.new_bookmark.connect(self.bookmarks_widget.create_requested)
         self.web_view.toggle_inspector.connect(self.toggle_inspector)
         self.web_view.toggle_lookup.connect(self.toggle_lookup)
         self.web_view.quit.connect(self.quit)
@@ -176,7 +181,11 @@ class EbookViewer(MainWindow):
         self.web_view.show_error.connect(self.show_error)
         self.web_view.print_book.connect(self.print_book, type=Qt.QueuedConnection)
         self.web_view.reset_interface.connect(self.reset_interface, type=Qt.QueuedConnection)
+        self.web_view.quit.connect(self.quit, type=Qt.QueuedConnection)
         self.web_view.shortcuts_changed.connect(self.shortcuts_changed)
+        self.web_view.scrollbar_context_menu.connect(self.scrollbar_context_menu)
+        self.web_view.close_prep_finished.connect(self.close_prep_finished)
+        self.web_view.highlights_changed.connect(self.highlights_changed)
         self.actions_toolbar.initialize(self.web_view, self.search_dock.toggleViewAction())
         self.setCentralWidget(self.web_view)
         self.loading_overlay = LoadingOverlay(self)
@@ -192,14 +201,40 @@ class EbookViewer(MainWindow):
             rmap[v].append(k)
         self.actions_toolbar.set_tooltips(rmap)
 
-    def toggle_inspector(self):
-        visible = self.inspector_dock.toggleViewAction().isChecked()
-        self.inspector_dock.setVisible(not visible)
-
     def resizeEvent(self, ev):
         self.loading_overlay.resize(self.size())
         return MainWindow.resizeEvent(self, ev)
 
+    def scrollbar_context_menu(self, x, y, frac):
+        m = QMenu(self)
+        amap = {}
+
+        def a(text, name):
+            m.addAction(text)
+            amap[text] = name
+
+        a(_('Scroll here'), 'here')
+        m.addSeparator()
+        a(_('Start of book'), 'start_of_book')
+        a(_('End of book'), 'end_of_book')
+        m.addSeparator()
+        a(_('Previous section'), 'previous_section')
+        a(_('Next section'), 'next_section')
+        m.addSeparator()
+        a(_('Start of current file'), 'start_of_file')
+        a(_('End of current file'), 'end_of_file')
+        m.addSeparator()
+        a(_('Hide this scrollbar'), 'toggle_scrollbar')
+
+        q = m.exec_(QCursor.pos())
+        if not q:
+            return
+        q = amap[q.text()]
+        if q == 'here':
+            self.web_view.goto_frac(frac)
+        else:
+            self.web_view.trigger_shortcut(q)
+
     # IPC {{{
     def handle_commandline_arg(self, arg):
         if arg:
@@ -245,6 +280,10 @@ class EbookViewer(MainWindow):
 
     # Docks (ToC, Bookmarks, Lookup, etc.) {{{
 
+    def toggle_inspector(self):
+        visible = self.inspector_dock.toggleViewAction().isChecked()
+        self.inspector_dock.setVisible(not visible)
+
     def toggle_toc(self):
         self.toc_dock.setVisible(not self.toc_dock.isVisible())
 
@@ -268,6 +307,14 @@ class EbookViewer(MainWindow):
         else:
             self.bookmarks_widget.bookmarks_list.setFocus(Qt.OtherFocusReason)
 
+    def toggle_highlights(self):
+        is_visible = self.highlights_dock.isVisible()
+        self.highlights_dock.setVisible(not is_visible)
+        if is_visible:
+            self.web_view.setFocus(Qt.OtherFocusReason)
+        else:
+            self.highlights_widget.focus()
+
     def toggle_lookup(self):
         self.lookup_dock.setVisible(not self.lookup_dock.isVisible())
 
@@ -457,7 +504,12 @@ class EbookViewer(MainWindow):
                 initial_position = {'type': 'ref', 'data': open_at[len('ref:'):]}
             elif is_float(open_at):
                 initial_position = {'type': 'bookpos', 'data': float(open_at)}
-        self.web_view.start_book_load(initial_position=initial_position)
+        highlights = self.current_book_data['annotations_map']['highlight']
+        self.highlights_widget.load(highlights)
+        self.web_view.start_book_load(
+            initial_position=initial_position,
+            highlights=list(map(serialize_annotation, highlights))
+        )
 
     def load_book_data(self):
         self.load_book_annotations()
@@ -508,6 +560,7 @@ class EbookViewer(MainWindow):
             return
         self.current_book_data['annotations_map']['last-read'] = [{
             'pos': cfi, 'pos_type': 'epubcfi', 'timestamp': utcnow()}]
+        self.save_pos_timer.start()
     # }}}
 
     # State serialization {{{
@@ -526,6 +579,15 @@ class EbookViewer(MainWindow):
                 save_annots_to_epub(path, annots)
                 update_book(path, before_stat, {'calibre-book-annotations.json': annots})
 
+    def highlights_changed(self, highlights):
+        if not self.current_book_data:
+            return
+        for h in highlights:
+            h['timestamp'] = parse_iso8601(h['timestamp'], assume_utc=True)
+        amap = self.current_book_data['annotations_map']
+        amap['highlight'] = highlights
+        self.save_annotations()
+
     def save_state(self):
         with vprefs:
             vprefs['main_window_state'] = bytearray(self.saveState(self.MAIN_WINDOW_STATE_VERSION))
@@ -536,6 +598,8 @@ class EbookViewer(MainWindow):
         geom = vprefs['main_window_geometry']
         if geom and get_session_pref('remember_window_geometry', default=False):
             QApplication.instance().safe_restore_geometry(self, geom)
+        else:
+            QApplication.instance().ensure_window_on_screen(self)
         if state:
             self.restoreState(state, self.MAIN_WINDOW_STATE_VERSION)
             self.inspector_dock.setVisible(False)
@@ -543,7 +607,24 @@ class EbookViewer(MainWindow):
     def quit(self):
         self.close()
 
+    def force_close(self):
+        if not self.close_forced:
+            self.close_forced = True
+            self.quit()
+
+    def close_prep_finished(self, cfi):
+        if cfi:
+            self.cfi_changed(cfi)
+        self.force_close()
+
     def closeEvent(self, ev):
+        if self.current_book_data and self.web_view.view_is_ready and not self.close_forced:
+            ev.ignore()
+            if not self.shutting_down:
+                self.shutting_down = True
+                QTimer.singleShot(2000, self.force_close)
+                self.web_view.prepare_for_close()
+            return
         self.shutting_down = True
         self.search_widget.shutdown()
         try:
